#
# This file is part of Zaplet
# Copyright 1999 Adam Feuer <adamf@pobox.com>
#
# Zaplet is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Zaplet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Zaplet; see the file COPYING-Zaplet. If not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#

# zaplet classes
# for use with adzapper filtering proxy

# Costas:
# Minor changes to make it work better with CTC's new network engine

RCS_ID =  '$Id: zaplet_engine.py,v 1.2 1999/09/07 06:44:29 adamf Exp $'

import os
import re
import string
import posixpath
import sys
import urlparse
import socket

# adzapper modules
import mdebug

numeric_host = re.compile ('[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+')
numeric_network = re.compile ('[0-9]+\.([0-9]+\.)?([0-9]+\.)?([[0-9]+\*])?')

# returns true if string is enclosed in quotes
def quoted_string(arg):
    l = len(arg)
    if l > 1:
        if arg[0] == "\"" and arg[l-1] == "\"":
            return 1
        elif arg[0] == "\'" and arg[l-1] == "\'":
            return 1
        return 0

# returns true of string is enclosed in '/'
def regex_string(arg):
    l = len(arg)
    if l > 1:
        if arg[0] == '/' and arg[l-1] == '/':
            return 1
    return 0

# cut off the first and last characters of the string
def trim(arg):
    l = len(arg)
    if l > 2:
        return arg[1:l-1]
    else:
        return ""



# data structure object to hold zaplet
class zaplet:
    def __init__(self):
        # zaplet version defaults to 0.1
        self.majorversion = None
        self.minorversion = None
        self.host = None
        self.allow = ''
        self.block = ''
        self.lines = []




# this does all the work--
# reads in zaplets, and determines whether URLs should be blocked.
class zaplet_engine:
    # initialize the zaplet_engine object
    # build the {site : block_regex} dictionary

    def __init__(self, zDir):
        self.zDir = zDir

        # dictionary of (site, zaplet object instance) pairs
        self.zaplet_list = { }

        self.readZaplets(self.zDir)



    # read Zaplets from the repositories
    # note: zDir must have a trailing '/' 
    def readZaplets(self,zDir):
        mdebug.mdebug(1,"Initializing zaplet engine...")
        cpairs = []  # list of (site, conf file) pairs

        # make the list of zaplets

        if (zDir[0] == '/'):
            # it is an absolute path
            pass
        else:
            # not absolute path to zaplet directory-- add cwd to it
            self.zDir = posixpath.join(os.getcwd(),zDir)
        mdebug.mdebug(1, "zaplet_engine, zaplet dir: %s" % self.zDir)

        zdirlist = os.listdir(self.zDir)
        mdebug.mdebug(1, "zaplet_engine, zdirlist:")
        mdebug.mdebug(1, zdirlist)

        zapfile =  re.compile('^.*\.zap')
        dotfile =  re.compile('^\.')
        conffile = re.compile('.*\.conf$')
        zlist = []
        for file in zdirlist:
            if posixpath.isfile(posixpath.join(self.zDir,file)):
                if zapfile.search(file) and not dotfile.search(file) and not conffile.search(file):
                    zlist.append(file)
        mdebug.mdebug(1,"processed source directory: ")
        mdebug.mdebug(1,zlist)

        # now go thru list of zaplets, read them in and compile them.
        for zfile in zlist:
            mdebug.mdebug(1,"site: " + zfile + "        zaplet file: " + zfile)
            zl = self.get_zaplet(posixpath.join(self.zDir,zfile))

            # if we got a list of zaplets back...
            if len(zl) > 0:
                # add each of the zaplets to the dictionary
                for zlitem in zl:
                    self.zaplet_list[zlitem.host] = zlitem

        mdebug.mdebug(1,self.zaplet_list.keys())



    # given the name of a zapletfile, read it in, split any zaplets into separate lists of lines,
    # and call make_zaplet() on each list of lines.
    #  note: zapletfile must be an absolute path to the zaplet file
    def get_zaplet(self,zapletfile):
        # make the list of lists
        zaplist = []

        try:
            mdebug.mdebug(1," opening zaplet : " + zapletfile)
            cf = open(zapletfile, "r")
        except:
            sys.stderr.write("Cannot open zaplet " + zapletfile + "\n")
            return None

        lines = cf.readlines()
        cf.close()

        in_zaplet = 0
        for line in lines:
            if in_zaplet:
                zlines.append(line)
                if re.search('^\s*</zaplet>',line):
                    in_zaplet = 0
                    zaplist.append(zlines)
                    mdebug.mdebug(1,zlines)
            else: # not in_zaplet
                if re.search('^\s*<zaplet>',line):
                    in_zaplet = 1
                    zlines = []
                    zlines.append(line)

        zaplet_list = []
        for zlitem in zaplist:
            tmp = self.make_zaplet(zlitem)
            if tmp != None:
                zaplet_list.append(tmp)
            else:
                mdebug.mdebug(1,"%s: zaplet not used." % zapletfile)

        return zaplet_list



    # given the list of lines that make a zaplet, parse the lines and return a zaplet object
    #   complete with compiled regexes
    # returns None if zaplet is invalid.
    def make_zaplet(self,zapletlines):
        # make a new zaplet
        z = zaplet()

        block_re_str = '' # the block regex that will be built for this zaplet
        allow_re_str = '' # the allow regex that will be built for this zaplet

        z.lines = zapletlines

        in_zaplet = 0
        for line in z.lines:
            if in_zaplet:
                line = string.strip(line)
                if (len(line) == 0) or re.search("\s*#", line):
                    continue
                if string.find(line, '#'):
                    ltmp = string.split(line, '#')
                    line = ltmp[0]
                tokens = re.split('\s+', line)
                command = string.lower(tokens[0])
                if (command == 'version'):
                    vstr = tokens[1]
                    vlist = string.split(vstr,'.')
                    z.majorversion = string.atoi(vlist[0])
                    z.minorversion = string.atoi(vlist[1])
                    mdebug.mdebug(1,'zaplet version: %i.%i' % (z.majorversion, z.minorversion))
                elif (command == 'host'):
                    z.host = tokens[1]
                    mdebug.mdebug(1,'zaplet host: %s' % z.host)
                elif (command == 'block'):
                    if len(tokens) > 1:
                        arg = tokens[1]
                    else:
                        arg = ""
                    if regex_string(arg):  
                        # regular expression - don't escape
                        # trim slashes and add to regex
                        mdebug.mdebug (2,("allow regular expression: ",arg))
                        re_add = trim(arg)
                        mdebug.mdebug(2,("block trimmed regex: ", re_add))
                        block_re_str =  block_re_str + '|' + re_add 
                    elif quoted_string(arg):
                        # literal string - trim quotes & escape
                        mdebug.mdebug (2,("block quoted string: ",arg))
                        re_add = trim(arg)
                        mdebug.mdebug(2, ("block trimmed string: ", re_add))
                        re_add = re.escape(re_add)
                        block_re_str = block_re_str + '|' + re_add
                    else:   # literal string - need to escape it.
                        re_add = re.escape(arg)
                        block_re_str = block_re_str + '|' + re_add
                        mdebug.mdebug(2, ("block literal string: ", re_add))
                elif (command == 'allow'):
                    if len(tokens) > 1:
                        mdebug.mdebug(2, "tokens[1]: %s" % tokens[1])
                        arg = tokens[1]
                    else:
                        arg = ""
                    if regex_string(arg):  
                        # regular expression - don't escape
                        # trim slashes and add to regex
                        mdebug.mdebug (2,"allow regular expression: %s" % arg)
                        re_add = trim(arg)
                        mdebug.mdebug(2,("allow trimmed regex: ", re_add))
                        allow_re_str =  allow_re_str + '|' + re_add 
                    elif quoted_string(arg):
                        # literal string - trim quotes & escape
                        mdebug.mdebug (2,("allow quoted string: ",arg))
                        re_add = trim(arg)
                        mdebug.mdebug(2, ("allow trimmed string: ", re_add))
                        re_add = re.escape(re_add)
                        allow_re_str = allow_re_str + '|' + re_add
                    else:   # literal string - need to escape it.
                        re_add = re.escape(arg)
                        allow_re_str = allow_re_str + '|' + re_add
                        mdebug.mdebug(2, ("allow literal string: ", re_add))

                elif (command == '</zaplet>'):
                    in_zaplet = 0
                    continue
                else:
                    mdebug.mdebug(1,"unknown command: " + command)
                    pass
            else: # not in_zaplet
                if re.search('^<zaplet>',line):
                    in_zaplet = 1
                    continue

        # if there is no version statement, don't use the zaplet--
        #   version statements are now required!
        if z.majorversion != None and z.minorversion != None:
            mdebug.mdebug(1,"z.version: %d.%d " % (z.majorversion,z.minorversion))
        else:
            mdebug.mdebug(1,"zaplet not used; no version statement!")
            return None

        # if there is no host statement, don't use the zaplet--
        #   host statements are now required
        if z.host != None:
            mdebug.mdebug(1,"z.host: %s " % z.host)
        else:
            mdebug.mdebug(1,"zaplet not used; no host statement!")
            return None


        # chop off initial |
        allow_re_str = allow_re_str[1:]
        block_re_str = block_re_str[1:]

        mdebug.mdebug(1,"allow_re_str: %s" % allow_re_str)
        mdebug.mdebug(1,"block_re_str: %s" % block_re_str)

        # compile regexes
        if allow_re_str == "":
            allow_re_str = "^$"
        z.allow = allow_re_str
        z.allow_re = re.compile(z.allow)

        z.block = block_re_str
        z.block_re = re.compile(z.block)

        return z



    # check to see whether a named host is blocked -
    # uses zaplet database, searchs from most specific to least specific
    # (whole dns name www.foo.bar.com, then foo.bar.com, then bar.com, then .com)
    # returns 1 if url should be blocked, 0 if not.

    def checkNamedHost(self,host,url):
        srv_fields = string.split(host,'.')
        #mdebug.mdebug(1,srv_fields)
        while (srv_fields != []):
            tmps = string.join(srv_fields,'.');
            if (self.zaplet_list.has_key(tmps)):
                # we are on the site-- check for allows first
                if (self.zaplet_list[tmps].allow_re.search(url)):
                    # it's specifically allowed, so let it thru
                    mdebug.mdebug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                    return 0
                # ok, it's not specifically allowed, now check for blocks
                elif (self.zaplet_list[tmps].block_re.search(url)):
                    mdebug.mdebug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                    return 1
                # it's not specifically blocked, so let it thru.
                else:
                    # return-- we don't want to continue searching
                    return 0
            del srv_fields[0]
        return 0



    # check to see whether a numeric host is blocked -
    # uses zaplet database, searchs from most specific to least specific
    # (whole ip number a.b.c.d, then network a.b.c, then net a.b, and then net a)
    # returns 1 if url should be blocked, 0 if not.

    def checkNumericHost(self,host,url):
        srv_fields = string.split(host,'.')
        mdebug.mdebug(1,"numeric host ip: %s" % srv_fields)
        lastfield = len(srv_fields)
        mdebug.mdebug(1,"numeric host ip fields: %d" % lastfield)
        for i in range(0,lastfield):
            tmps = string.join(srv_fields[0:lastfield-i],'.');
            mdebug.mdebug(1,"numeric host- iteration ip: %s" % tmps)
            if (self.zaplet_list.has_key(tmps)):
                # we are on the site-- check for allows first
                if (self.zaplet_list[tmps].allow_re.search(url)):
                    # it's specifically allowed, so let it thru
                    mdebug.mdebug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                    return 0
                # ok, it's not specifically allowed, now check for blocks
                elif (self.zaplet_list[tmps].block_re.search(url)):
                    mdebug.mdebug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                    return 1
                # it's not specifically blocked, so let it thru.
                else:
                    # return-- we don't want to continue searching
                    return 0
                # default zaplet

        tmps = 'default-numeric'
        mdebug.mdebug(1,"checking default-numeric...")
        if (self.zaplet_list.has_key(tmps)):
            if (self.zaplet_list[tmps].allow_re.search(url)):
                # it's specifically allowed, so let it thru
                mdebug.mdebug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                return 0
            # ok, it's not specifically allowed, now check for blocks
            elif (self.zaplet_list[tmps].block_re.search(url)):
                mdebug.mdebug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                return 1
        return 0


 

    # determine whether a URL should be zapped or not
    # returns 1 if it should be blocked, 0 otherwise
    def urlblocker(self, rem_host, url):
        # does the remote host match a zaplet in the database?
        if numeric_host.match(rem_host) == None:
            # has DNS name
            mdebug.mdebug(1, "named host.")
            if self.checkNamedHost(rem_host,url) == 1:
                # yes
                return 1
        else:
            # is a numeric host
            mdebug.mdebug(1, "numeric host.")
            if self.checkNumericHost(rem_host,url) == 1:
                return 1

        # default zaplet
        mdebug.mdebug(1,"checking default...")
        tmps = 'default'
        if (self.zaplet_list.has_key(tmps)):
            mdebug.mdebug(2,"zap: has default zaplet...")
            if (self.zaplet_list[tmps].allow_re.search(url)):
                # it's specifically allowed, so let it thru
                mdebug.mdebug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                return 0
            # ok, it's not specifically allowed, now check for blocks
            elif (self.zaplet_list[tmps].block_re.search(url)):
                mdebug.mdebug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                return 1
        mdebug.mdebug(2,"zap: no allows and no blocks. (%s)" % url)
        return 0


